# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add_storage_driver_to_backups_table

Revision ID: 7ee6154548a6
Revises: cee1bcba3541
Create Date: 2024-06-18 16:14:38.561592

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy import Column
from sqlalchemy import Text, String
from sqlalchemy.sql import table, column


# revision identifiers, used by Alembic.
revision: str = '7ee6154548a6'
down_revision: Union[str, None] = 'cee1bcba3541'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column("backups", Column('storage_driver', Text(),
                                    nullable=True))
    backups_tables = table(
        "backups",
        column("storage_driver", String)
    )
    op.execute(
        backups_tables.update()
        .where(backups_tables.c.storage_driver.is_(None))
        .values({"storage_driver": "swift"})
    )


def downgrade() -> None:
    op.drop_column("backups", 'storage_driver')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from unittest.mock import MagicMock, Mock, PropertyMock, patch

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF.backup_encryption_key = None
CONF.backup_id = "backup_unittest"
CONF.db_user = "db_user"
CONF.db_password = "db_password"
CONF.db_host = "db_host"

driver_mapping = {
    'innobackupex': 'backup.drivers.innobackupex.InnoBackupEx',
    'innobackupex_inc': 'backup.drivers.innobackupex.InnoBackupExIncremental',
    'mariabackup': 'backup.drivers.mariabackup.MariaBackup',
    'mariabackup_inc': 'backup.drivers.mariabackup.MariaBackupIncremental',
    'pg_basebackup': 'backup.drivers.postgres.PgBasebackup',
    'pg_basebackup_inc': 'backup.drivers.postgres.PgBasebackupIncremental',
    'xtrabackup': 'backup.drivers.xtrabackup.XtraBackup',
    'xtrabackup_inc': 'backup.drivers.xtrabackup.XtraBackupIncremental'
}

from backup.drivers.xtrabackup import XtraBackupException


class TestXtraBackup(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['xtrabackup'])
        self.params = {
            'db_datadir': '/var/lib/mysql/data'
        }
        # assertions
        self.assertIsNotNone(self.runner_cls)

    def tearDown(self):
        pass

    def test_instance(self):
        '''Check instance'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertIsNotNone(runner)

    def test_cmd(self):
        '''Check cmd property'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        cmd = ('xtrabackup --backup --stream=xbstream --parallel=2 '
               '--datadir=%(datadir)s --user=%(user)s '
               '--password=%(password)s --host=%(host)s'
               % {
                   'datadir': runner.datadir,
                   'user': CONF.db_user,
                   'password': CONF.db_password,
                   'host': CONF.db_host}
               )
        self.assertEqual(runner.cmd, cmd)

    def test_check_restore_process(self):
        '''Check manifest'''
        # call the method
        runner = self.runner_cls(**self.params)
        runner.process = MagicMock()
        returncode = PropertyMock(return_value=0)
        type(runner.process).returncode = returncode

        # assertions
        self.assertEqual(runner.check_restore_process(), True)

    def test_post_restore(self):
        '''Check manifest'''
        runner = self.runner_cls(**self.params)
        mock = Mock(side_effect=XtraBackupException(
            'Prepare did not complete successfully'))
        runner.post_restore = mock

        # call the method
        with self.assertRaises(
            XtraBackupException,
                msg="Prepare did not complete successfully"):
            runner.post_restore()


# Manually import XtraBackupIncremental to prevent from running
# xtrabackup --version when calling the TestXtraBackupIncremental
# constructor
from backup.drivers.xtrabackup import XtraBackupIncremental


class TestXtraBackupIncremental(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['xtrabackup_inc'])
        self.params = {
            'lsn': '1234567890',
            'parent_location': '',
            'parent_checksum': '',
        }
        self.metadata = {
            'lsn': '1234567890',
            'parent_location': 'https://example.com/location',
            'parent_checksum': 'f1508ecf362a364c5aae008b4b5a9cb9',
        }

    def tearDown(self):
        pass

    def test_instance(self):
        '''Check instance and add_incremental_opts'''
        # call the method
        with patch(
            'backup.drivers.xtrabackup.XtraBackupIncremental.'
            'add_incremental_opts', new_callable=PropertyMock
        ) as XtraBackupIncremental_add_incremental_opts:
            XtraBackupIncremental_add_incremental_opts.return_value = True
            runner = XtraBackupIncremental(**self.params)
            # assertions
            self.assertIsNotNone(runner)

    def test_cmd(self):
        '''Check cmd property'''
        # call the method
        with patch(
            'backup.drivers.xtrabackup.XtraBackupIncremental.'
            'add_incremental_opts', new_callable=PropertyMock
        ) as XtraBackupIncremental_add_incremental_opts:
            XtraBackupIncremental_add_incremental_opts.return_value = True
            runner = XtraBackupIncremental(**self.params)

            # assertions
            self.assertIsNotNone(runner)
            # assertions
            cmd = (f'xtrabackup --backup --stream=xbstream '
                   f'--incremental-lsn=%(lsn)s '
                   f'--datadir={runner.datadir} {runner.user_and_pass}')
            if runner.add_incremental_opts:
                cmd = '{} --incremental'.format(cmd)
            self.assertEqual(runner.cmd, cmd)

    def test_get_metadata(self):
        '''Check get_metadata'''
        with patch(
            'backup.drivers.xtrabackup.XtraBackupIncremental.'
            'add_incremental_opts', new_callable=PropertyMock
        ) as XtraBackupIncremental_add_incremental_opts:
            XtraBackupIncremental_add_incremental_opts.return_value = True
            runner = XtraBackupIncremental(**self.params)
            runner.get_metadata = MagicMock(return_value=self.metadata)

            # assertions
            self.assertIsNotNone(runner)
            ret = runner.get_metadata()
            self.assertEqual(ret, self.metadata)

    def test_run_restore(self):
        '''Check run_restore'''
        with patch(
            'backup.drivers.xtrabackup.XtraBackupIncremental.'
            'add_incremental_opts', new_callable=PropertyMock
        ) as XtraBackupIncremental_add_incremental_opts:
            XtraBackupIncremental_add_incremental_opts.return_value = True
            runner = XtraBackupIncremental(**self.params)
            length = 10
            runner.incremental_restore = MagicMock(return_value=length)
            runner.restore_content_length = length
            # call the method
            ret = runner.run_restore()
            # assertions
            self.assertEqual(ret, length)


if __name__ == '__main__':
    unittest.main()



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os

from oslo_log import log as logging

from trove.common import cfg
from trove.common import exception
from trove.guestagent.common import operating_system
from trove.guestagent.strategies.replication import mysql_base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class MariaDBGTIDReplication(mysql_base.MysqlReplicationBase):
    """MariaDB Replication coordinated by GTIDs."""

    def get_replica_context(self, service, adm):
        """Get replication information as master."""
        master_info = super(MariaDBGTIDReplication, self).get_replica_context(
            service, adm)

        replica_conf = master_info['replica_conf']
        get_pos_cmd = 'SELECT @@global.gtid_binlog_pos;'
        gtid_pos = service.execute_sql(get_pos_cmd).first()[0]
        LOG.debug('gtid_binlog_pos: %s', gtid_pos)
        replica_conf['log_position']['gtid_pos'] = gtid_pos

        return master_info

    def read_last_master_gtid(self, service):
        # mariadb > 10.5  uses mariadb_backup_binlog_info instead.
        xtrabackup_info_file = ('%s/xtrabackup_binlog_info'
                                % service.get_data_dir())
        mariabackup_info_file = ('%s/mariadb_backup_binlog_info'
                                 % service.get_data_dir())

        if os.path.exists(mariabackup_info_file):
            INFO_FILE = mariabackup_info_file
            LOG.info("Using mariabackup_info_file")
        elif os.path.exists(xtrabackup_info_file):
            INFO_FILE = xtrabackup_info_file
            LOG.info("Using xtrabackup_binlog_info")
        else:
            # Handle the case where neither file exists
            LOG.error("Neither xtrabackup_binlog_info nor "
                      "mariadb_backup_binlog_info found.")
            raise exception.UnableToDetermineLastMasterGTID(
                binlog_file="xtrabackup_binlog_info or"
                            "mariadb_backup_binlog_info")

        operating_system.chmod(INFO_FILE,
                               operating_system.FileMode.ADD_READ_ALL,
                               as_root=True)

        LOG.info("Reading last master GTID from %s", INFO_FILE)
        try:
            with open(INFO_FILE, 'r') as f:
                content = f.read()
                LOG.debug('Content in %s: "%s"', INFO_FILE, content)
                ret = content.strip().split('\t')
                return ret[2] if len(ret) == 3 else ''
        except Exception as ex:
            LOG.error('Failed to read last master GTID, error: %s', str(ex))
            raise exception.UnableToDetermineLastMasterGTID(
                binlog_file=INFO_FILE) from ex

    def connect_to_master(self, service, master_info):
        replica_conf = master_info['replica_conf']
        last_gtid = ''

        if 'dataset' in master_info:
            # This will happen when initial replication is set up.
            last_gtid = self.read_last_master_gtid(service)
            set_gtid_cmd = "SET GLOBAL gtid_slave_pos='%s';" % last_gtid
            service.execute_sql(set_gtid_cmd)

        change_master_cmd = (
            "CHANGE MASTER TO "
            "MASTER_HOST='%(host)s', "
            "MASTER_PORT=%(port)s, "
            "MASTER_USER='%(user)s', "
            "MASTER_PASSWORD='%(password)s', "
            "MASTER_CONNECT_RETRY=15, "
            "MASTER_USE_GTID=slave_pos" %
            {
                'host': master_info['master']['host'],
                'port': master_info['master']['port'],
                'user': replica_conf['replication_user']['name'],
                'password': replica_conf['replication_user']['password']
            })
        service.execute_sql(change_master_cmd)

        service.start_slave()



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import abc
import json
import os
import re

from pyroute2 import IPRoute

from trove.common import cfg
from trove.common import constants
from trove.common import pagination
from trove.common import utils
from trove.guestagent.common import operating_system

CONF = cfg.CONF


def update_dict(updates, target):
    """Recursively update a target dictionary with given updates.

    Updates are provided as a dictionary of key-value pairs
    where a value can also be a nested dictionary in which case
    its key is treated as a sub-section of the outer key.
    If a list value is encountered the update is applied
    iteratively on all its items.

    :returns:    Will always return a dictionary of results (may be empty).
    """
    if target is None:
        target = {}

    if isinstance(target, list):
        for index, item in enumerate(target):
            target[index] = update_dict(updates, item)
        return target

    if updates is not None:
        for k, v in updates.items():
            if isinstance(v, abc.Mapping):
                target[k] = update_dict(v, target.get(k, {}))
            else:
                target[k] = updates[k]

    return target


def expand_dict(target, namespace_sep='.'):
    """Expand a flat dict to a nested one.
    This is an inverse of 'flatten_dict'.

    :seealso: flatten_dict
    """
    nested = {}
    for k, v in target.items():
        sub = nested
        keys = k.split(namespace_sep)
        for key in keys[:-1]:
            sub = sub.setdefault(key, {})
        sub[keys[-1]] = v

    return nested


def flatten_dict(target, namespace_sep='.'):
    """Flatten a nested dict.
    Return a one-level dict with all sub-level keys joined by a namespace
    separator.

    The following nested dict:
    {'ns1': {'ns2a': {'ns3a': True, 'ns3b': False}, 'ns2b': 10}}

    would be flattened to:
    {'ns1.ns2a.ns3a': True, 'ns1.ns2a.ns3b': False, 'ns1.ns2b': 10}
    """
    def flatten(target, keys, namespace_sep):
        flattened = {}
        if isinstance(target, abc.Mapping):
            for k, v in target.items():
                flattened.update(
                    flatten(v, keys + [k], namespace_sep))
        else:
            ns = namespace_sep.join(keys)
            flattened[ns] = target

        return flattened

    return flatten(target, [], namespace_sep)


def build_file_path(base_dir, base_name, *extensions):
    """Build a path to a file in a given directory.
    The file may have an extension(s).

    :returns:    Path such as: 'base_dir/base_name.ext1.ext2.ext3'
    """
    file_name = os.extsep.join([base_name] + list(extensions))
    return os.path.expanduser(os.path.join(base_dir, file_name))


def to_bytes(value):
    """Convert numbers with a byte suffix to bytes.
    """
    if isinstance(value, str):
        pattern = re.compile(r'^(\d+)([K,M,G]{1})$')
        match = pattern.match(value)
        if match:
            value = match.group(1)
            suffix = match.group(2)
            factor = {
                'K': 1024,
                'M': 1024 ** 2,
                'G': 1024 ** 3,
            }[suffix]

            return int(round(factor * float(value)))

    return value


def paginate_list(li, limit=None, marker=None, include_marker=False):
    """Paginate a list of objects based on the name attribute.
    :returns:           Page sublist and a marker (name of the last item).
    """
    return pagination.paginate_object_list(
        li, 'name', limit=limit, marker=marker, include_marker=include_marker)


def serialize_list(li, limit=None, marker=None, include_marker=False):
    """
    Paginate (by name) and serialize a given object list.
    :returns:           A serialized and paginated version of a given list.
    """
    page, next_name = paginate_list(li, limit=limit, marker=marker,
                                    include_marker=include_marker)
    return [item.serialize() for item in page], next_name


def get_filesystem_volume_stats(fs_path):
    try:
        stats = os.statvfs(fs_path)
    except OSError:
        raise RuntimeError("Filesystem not found (%s)" % fs_path)

    total = stats.f_blocks * stats.f_bsize
    free = stats.f_bfree * stats.f_bsize
    # return the size in GB
    used_gb = utils.to_gb(total - free)
    total_gb = utils.to_gb(total)

    output = {
        'block_size': stats.f_bsize,
        'total_blocks': stats.f_blocks,
        'free_blocks': stats.f_bfree,
        'total': total_gb,
        'free': free,
        'used': used_gb
    }
    return output


def get_conf_dir():
    """Get the config directory for the database related settings.

    For now, the files inside the config dir are mainly for instance rebuild.
    """
    mount_point = CONF.get(CONF.datastore_manager).mount_point
    conf_dir = os.path.join(mount_point, 'conf.d')
    if not operating_system.exists(conf_dir, is_directory=True, as_root=True):
        operating_system.ensure_directory(conf_dir, as_root=True)

    return conf_dir


def disable_user_defined_port():
    with open(constants.ETH1_CONFIG_PATH) as fd:
        eth1_config = json.load(fd)
    ipr = IPRoute()
    ifaces = ipr.get_links(address=eth1_config.get("mac_address"))
    if not ifaces:
        return
    ifname = ifaces[0].get_attr('IFLA_IFNAME')
    operating_system.execute_shell_cmd(f"ip link set {ifname} down", [],
                                       shell=True,
                                       as_root=True)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from unittest.mock import MagicMock, PropertyMock

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF.backup_encryption_key = None
CONF.backup_id = "backup_unittest"
CONF.db_user = "db_user"
CONF.db_password = "db_password"
CONF.db_host = "db_host"

driver_mapping = {
    'innobackupex': 'backup.drivers.innobackupex.InnoBackupEx',
    'innobackupex_inc': 'backup.drivers.innobackupex.InnoBackupExIncremental',
    'mariabackup': 'backup.drivers.mariabackup.MariaBackup',
    'mariabackup_inc': 'backup.drivers.mariabackup.MariaBackupIncremental',
    'pg_basebackup': 'backup.drivers.postgres.PgBasebackup',
    'pg_basebackup_inc': 'backup.drivers.postgres.PgBasebackupIncremental',
    'xtrabackup': 'backup.drivers.xtrabackup.XtraBackup',
    'xtrabackup_inc': 'backup.drivers.xtrabackup.XtraBackupIncremental'
}


class TestMariaBackup(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['mariabackup'])
        self.params = {}

        # assertions
        self.assertIsNotNone(self.runner_cls)

    def tearDown(self):
        pass

    def test_instance(self):
        '''Check instance'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        self.assertIsNotNone(runner)

    def test_cmd(self):
        '''Check cmd property'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        cmd = ("mariadb-backup --backup --stream=xbstream {}".format(
            runner.user_and_pass))
        self.assertEqual(runner.cmd, cmd)

    def test_check_restore_process(self):
        '''Check manifest'''
        runner = self.runner_cls(**self.params)
        runner.process = MagicMock()
        returncode = PropertyMock(return_value=0)
        type(runner.process).returncode = returncode

        # call the method
        self.assertEqual(runner.check_restore_process(), True)


class TestMariaBackupIncremental(unittest.TestCase):
    def setUp(self):
        self.runner_cls = importutils.import_class(
            driver_mapping['mariabackup_inc'])
        self.params = {
            'lsn': '1234567890',
            'incremental_dir': './'
        }
        self.metadata = {}

    def tearDown(self):
        pass

    def test_cmd(self):
        '''Check cmd property'''
        # call the method
        runner = self.runner_cls(**self.params)

        # assertions
        cmd = (
            'mariadb-backup --backup --stream=xbstream'
            ' --incremental-lsn=%(lsn)s ' +
            runner.user_and_pass
        )
        self.assertEqual(runner.cmd, cmd)

    def test_get_metadata(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        runner.get_metadata = MagicMock(return_value=self.metadata)

        # call the method
        ret = runner.get_metadata()

        # assertions
        self.assertEqual(ret, self.metadata)

    def test_run_restore(self):
        # prepare the test
        runner = self.runner_cls(**self.params)
        length = 10
        runner.incremental_restore = MagicMock(return_value=length)
        runner.restore_content_length = length

        # call the method
        ret = runner.run_restore()

        # assertions
        self.assertEqual(ret, length)


if __name__ == '__main__':
    unittest.main()
